"""
Classes and utilities for character range expressions that define sets of allowed or disallowed characters.
"""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from grammatica.constants import (
    ALWAYS_SAFE_CHARS,
    CHAR_ESCAPE_MAP,
    RANGE_ESCAPE_CHARS,
)
from grammatica.grammar.base import Grammar
from grammatica.grammar.string import String
from grammatica.utils import char_to_hex

if sys.version_info >= (3, 12):  # pragma: no cover
    from typing import override
else:  # pragma: no cover
    from typing_extensions import override

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Any


class CharRange(Grammar):
    """Grammar that defines a set of allowed or disallowed characters.

    Args:
        char_ranges (Iterable[tuple[str, str]]): Character ranges in the form of tuples (start, end).
            Each range is inclusive, meaning both start and end characters are included.
            For example, ('a', 'z') includes all lowercase letters from 'a' to 'z'.
        negate (bool, optional): Negate the character range. Defaults to False.

    Raises:
        ValueError: Empty character range provided.
        ValueError: Start or end of a character range is not a single character.
        ValueError: End character is less than start character in a character range.
    """

    __slots__: tuple[str, ...] = ("char_ranges", "negate")

    def __init__(
        self,
        char_ranges: Iterable[tuple[str, str]],
        negate: bool = False,
    ) -> None:
        super().__init__()

        unsorted_char_ranges = list(char_ranges)
        if len(unsorted_char_ranges) == 0:
            raise ValueError("char_ranges must not be empty")
        for start, end in unsorted_char_ranges:
            if len(start) != 1:
                raise ValueError("start must be a single character")
            if len(end) != 1:
                raise ValueError("end must be a single character")
            if end < start:
                raise ValueError("end must be greater than or equal to start")
        ords: set[int] = set()
        for cstart, cend in unsorted_char_ranges:
            ords.update(range(ord(cstart), ord(cend) + 1))
        ord_ranges = self._iter_ords_to_ord_ranges(ords)
        self.char_ranges: list[tuple[str, str]] = list(
            map(self._ord_range_to_char_range, ord_ranges)
        )
        """Character ranges in the form of tuples (start, end).
            Each range is inclusive, meaning both start and end characters are included.
            For example, ('a', 'z') includes all lowercase letters from 'a' to 'z'.
        """
        self.negate: bool = negate
        """Negate the character range."""

    def render(self, full: bool = True, wrap: bool = True) -> str | None:
        if len(self.char_ranges) == 0:
            return None
        # TODO: Should character ranges be validated before rendering? They're currently only validated upon construction.
        ords: set[int] = set()
        for cstart, cend in self.char_ranges:
            ords.update(range(ord(cstart), ord(cend) + 1))
        expr = "["
        if self.negate:
            expr += "^"
        for start, end in self._iter_ords_to_ord_ranges(ords):
            start_esc = self._escape(chr(start))
            if start == end:
                expr += start_esc
                continue
            end_esc = self._escape(chr(end))
            if end == start + 1:
                expr += f"{start_esc}{end_esc}"
            else:
                expr += f"{start_esc}-{end_esc}"
        expr += "]"
        return expr

    def simplify(self) -> CharRange | String | None:
        n = len(self.char_ranges)
        if n == 0:
            return None
        if (n == 1) and (self.char_ranges[0][0] == self.char_ranges[0][1]):
            return String(self.char_ranges[0][0])
        return CharRange(self.char_ranges.copy(), negate=self.negate)

    @staticmethod
    def _escape(char: str) -> str:
        """Escape a character for use in a character range.

        Args:
            char (str): Character to escape.

        Returns:
            str: Escaped character.
        """
        if char in RANGE_ESCAPE_CHARS:
            return "\\" + char
        if char in ALWAYS_SAFE_CHARS:
            return char
        if char in CHAR_ESCAPE_MAP:
            return CHAR_ESCAPE_MAP[char]
        return char_to_hex(char)

    @classmethod
    def from_chars(cls, chars: Iterable[str], negate: bool = False) -> CharRange:
        """Create an instance of :class:`grammatica.grammar.CharRange` from an iterable of characters.

        Note:
            Duplicate characters in the input are ignored.

        Args:
            chars (Iterable[str]): Characters to include in the character ranges.
            negate (bool, optional): Negate the character ranges. Defaults to False.

        Returns:
            CharRange: Instance created from the provided characters.

        Raises:
            ValueError: No characters provided.
        """
        return cls.from_ords(map(ord, chars), negate=negate)

    @staticmethod
    def _ord_range_to_char_range(ord_range: tuple[int, int]) -> tuple[str, str]:
        """Convert a range of ordinals to a range of characters.

        Args:
            ord_range (tuple[int, int]): Ordinal range in the form of (start, end).
                The range is inclusive, meaning both start and end ordinals are included.
                For example, (97, 122) includes all lowercase letters from 'a' to 'z'.

        Returns:
            tuple[str, str]: Character range corresponding to the provided ordinal range.

        Raises:
            ValueError: Invalid ordinal(s) in ordinal range.
        """
        return chr(ord_range[0]), chr(ord_range[1])

    @classmethod
    def from_ords(cls, ords: Iterable[int], negate: bool = False) -> CharRange:
        """Create an instance of :class:`grammatica.grammar.CharRange` from an iterable of ordinals.

        Note:
            Duplicate ordinals in the input are ignored.

        Args:
            ords (Iterable[int]): Ordinals to include in the character ranges.
            negate (bool, optional): Negate the character ranges. Defaults to False.

        Returns:
            CharRange: Instance created from the provided ordinals.

        Raises:
            ValueError: No ordinals provided.
        """
        ord_ranges = cls._iter_ords_to_ord_ranges(ords)
        char_ranges = map(cls._ord_range_to_char_range, ord_ranges)
        return cls(char_ranges, negate=negate)

    @staticmethod
    def _iter_ords_to_ord_ranges(ords: Iterable[int]) -> Iterator[tuple[int, int]]:
        """Generate ranges of consecutive ordinals from an iterable of ordinals.

        Args:
            ords (Iterable[int]): Ordinals to generate ordinal ranges from.

        Yields:
            tuple[int, int]: Ordinal range in the form of (start, end).
                The range is inclusive, meaning both start and end ordinals are included.
                For example, (97, 122) includes all lowercase letters from 'a' to 'z'.

        Raises:
            ValueError: No ordinals provided.
        """
        sorted_ords = sorted(set(ords))
        start = sorted_ords[0]
        end = start
        for i in range(1, len(sorted_ords)):
            ord_ = sorted_ords[i]
            if ord_ != end + 1:
                yield start, end
                start = ord_
            end = ord_
        yield start, end

    @override
    def as_string(self, indent: int | None = None) -> str:
        return super().as_string(indent=None)

    def attrs_dict(self) -> dict[str, Any]:
        return {
            "char_ranges": self.char_ranges,
            "negate": self.negate,
        }



import pytest

from grammatica.grammar import String
from grammatica.grammar.group import (
    And,
    Or,
    merge_adjacent_default_or_grammars,
)

try:
    from ..helpers import fmt_result
except ImportError:
    import sys
    from os import path

    sys.path.append(path.dirname(path.dirname(path.dirname(path.abspath(__file__)))))
    from helpers import fmt_result


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Empty Or",
            "grammar": Or([]),
            "expected": None,
        },
        {
            "description": "Empty Or with (0, 1) quantifier",
            "grammar": Or([], quantifier=(0, 1)),
            "expected": None,
        },
        {
            "description": "Or with subexpressions that all simplify to None returns None",
            "grammar": Or([String("")]),
            "expected": None,
        },
        {
            "description": "Or with single subexpression",
            "grammar": Or([String("a")]),
            "expected": String("a"),
        },
        {
            "description": "Or with single subexpression and (0, 1) quantifier",
            "grammar": Or([String("a")], quantifier=(0, 1)),
            "expected": And([String("a")], quantifier=(0, 1)),
        },
        {
            "description": "Or with consecutive String subexpressions",
            "grammar": Or([String("a"), String("b")]),
            "expected": Or([String("a"), String("b")]),
        },
        {
            "description": "Or with consecutive String subexpressions and (0, 1) quantifier",
            "grammar": Or([String("a"), String("b")], quantifier=(0, 1)),
            "expected": Or([String("a"), String("b")], quantifier=(0, 1)),
        },
        {
            "description": "Nested Or with consecutive String subexpressions",
            "grammar": Or(
                [
                    Or([String("a"), String("b")]),
                    Or([String("c"), String("d")]),
                ],
            ),
            "expected": Or([String("a"), String("b"), String("c"), String("d")]),
        },
        {
            "description": "Nested Or with (0, 1) quantifiers",
            "grammar": Or(
                [
                    Or([String("a")], quantifier=(0, 1)),
                ],
                quantifier=(0, 1),
            ),
            "expected": And([String("a")], quantifier=(0, 1)),
        },
        {
            "description": "Nested Or with consecutive String subexpressions and (0, 1) quantifiers",
            "grammar": Or(
                [
                    Or([String("a"), String("b")], quantifier=(0, 1)),
                    Or([String("c"), String("d")], quantifier=(0, 1)),
                ],
                quantifier=(0, 1),
            ),
            "expected": Or(
                [
                    Or([String("a"), String("b")], quantifier=(0, 1)),
                    Or([String("c"), String("d")], quantifier=(0, 1)),
                ],
                quantifier=(1, 1),
            ),
        },
        {
            "description": "Nested Or with consecutive String subexpressions and (0, 5) quantifiers",
            "grammar": Or(
                [
                    Or([String("a"), String("b")], quantifier=(0, 5)),
                    Or([String("c"), String("d")], quantifier=(0, 5)),
                ],
                quantifier=(0, 5),
            ),
            "expected": Or(
                [
                    Or([String("a"), String("b")], quantifier=(0, 5)),
                    Or([String("c"), String("d")], quantifier=(0, 5)),
                ],
                quantifier=(1, 5),
            ),
        },
        {
            "description": "Or with simple repeating subexpression",
            "grammar": Or([String("a"), String("a"), String("a")]),
            "expected": String("a"),
        },
        {
            "description": "Or with simple repeating subexpression and (0, 1) quantifier",
            "grammar": Or([String("a"), String("a"), String("a")], quantifier=(0, 1)),
            "expected": And([String("a")], quantifier=(0, 1)),
        },
        {
            "description": "Or with complex repeating subexpression",
            "grammar": Or(
                [
                    And([String("a"), String("b")]),
                    And([String("a"), String("b")]),
                    And([String("a"), String("b")]),
                ],
            ),
            "expected": String("ab"),
        },
        {
            "description": "Or with complex repeating subexpression and (0, 1) quantifiers",
            "grammar": Or(
                [
                    And([String("a"), String("b")], quantifier=(0, 1)),
                    And([String("a"), String("b")], quantifier=(0, 1)),
                    And([String("a"), String("b")], quantifier=(0, 1)),
                ],
                quantifier=(0, 1),
            ),
            "expected": And([String("ab")], quantifier=(0, 1)),
        },
    ],
)
def test_or_simplify(test_case):
    actual = test_case["grammar"].simplify()
    assert (actual == test_case["expected"]) and (
        (test_case["expected"] is not actual) or (test_case["expected"] is actual is None)
    ), "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(actual)!s}",
        )
    )


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Empty Or",
            "grammar": Or([]),
            "expected": None,
        },
        {
            "description": "Empty Or with (0, 1) quantifier",
            "grammar": Or([], quantifier=(0, 1)),
            "expected": None,
        },
        {
            "description": "Or with multiple subexpressions",
            "grammar": Or([String("a"), String("b")]),
            "expected": '("a" | "b")',
        },
        {
            "description": "Or with multiple subexpressions and (0, 1) quantifier",
            "grammar": Or([String("a"), String("b")], quantifier=(0, 1)),
            "expected": '("a" | "b")?',
        },
        {
            "description": "Or with multiple subexpressions and (0, None) quantifier",
            "grammar": Or([String("a"), String("b")], quantifier=(0, None)),
            "expected": '("a" | "b")*',
        },
        {
            "description": "Or with multiple subexpressions and (1, None) quantifier",
            "grammar": Or([String("a"), String("b")], quantifier=(1, None)),
            "expected": '("a" | "b")+',
        },
        {
            "description": "Or with multiple subexpressions and (1, 3) quantifier",
            "grammar": Or([String("a"), String("b")], quantifier=(1, 3)),
            "expected": '("a" | "b"){1,3}',
        },
        {
            "description": "Or with single subexpression",
            "grammar": Or([String("a")]),
            "expected": '"a"',
        },
        {
            "description": "Or with single subexpression and (0, 1) quantifier",
            "grammar": Or([String("a")], quantifier=(0, 1)),
            "expected": '"a"?',
        },
    ],
)
def test_or_render(test_case):
    actual = test_case["grammar"].render()
    assert (actual == test_case["expected"]) and (
        (test_case["expected"] is not actual) or (test_case["expected"] is actual is None)
    ), "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(actual)!s}",
        )
    )


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Empty Or",
            "grammar": Or([]),
            "expected": False,
        },
        {
            "description": "Or with single subexpression and default quantifier",
            "grammar": Or([String("a")]),
            "expected": False,
        },
        {
            "description": "Or with single subexpression and non-default quantifier",
            "grammar": Or([String("a")], quantifier=(0, 1)),
            "expected": False,
        },
        {
            "description": "Or with multiple subexpressions and default quantifier",
            "grammar": Or([String("a"), String("b")]),
            "expected": True,
        },
        {
            "description": "Or with multiple subexpressions and non-default quantifier",
            "grammar": Or([String("a"), String("b")], quantifier=(0, 1)),
            "expected": True,
        },
        {
            "description": "Nested Or with single subexpression and default quantifier",
            "grammar": Or([Or([String("a")])]),
            "expected": False,
        },
        {
            "description": "Nested Or with single subexpression and non-default quantifier",
            "grammar": Or([Or([String("a")])], quantifier=(0, 1)),
            "expected": False,
        },
    ],
)
def test_or_needs_wrapped(test_case):
    actual = test_case["grammar"].needs_wrapped()
    assert actual is test_case["expected"], "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(actual)!s}",
        )
    )


def test_or_attrs_dict():
    string_a = String("a")
    string_b = String("b")
    or_expr = Or([string_a, string_b], quantifier=(1, 2))
    expected = {
        "subexprs": [string_a, string_b],
        "quantifier": (1, 2),
    }
    actual = or_expr.attrs_dict()
    assert actual == expected
    for i in range(len(actual["subexprs"])):
        assert actual["subexprs"][i] is expected["subexprs"][i]


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Empty list",
            "subexprs": [],
            "expected": [],
        },
        {
            "description": "Single Or",
            "subexprs": [Or([String("a")])],
            "expected": [Or([String("a")])],
        },
        {
            "description": "Multiple adjacent Or",
            "subexprs": [Or([String("a")]), Or([String("b")]), Or([String("c")])],
            "expected": [Or([String("a"), String("b"), String("c")])],
        },
        {
            "description": "Mixed adjacent Or and other grammars",
            "subexprs": [
                Or([String("a")]),
                Or([String("b")]),
                And([String("c"), String("d")]),
                Or([String("e")]),
                Or([String("f")]),
            ],
            "expected": [
                Or([String("a"), String("b")]),
                And([String("c"), String("d")]),
                Or([String("e"), String("f")]),
            ],
        },
        {
            "description": "No adjacent Or",
            "subexprs": [
                String("a"),
                Or([String("b")]),
                And([String("c")]),
                String("d"),
            ],
            "expected": [
                String("a"),
                Or([String("b")]),
                And([String("c")]),
                String("d"),
            ],
        },
    ],
)
def test_merge_adjacent_default_or_grammars(test_case):
    subexprs = [expr.copy() for expr in test_case["subexprs"]]
    n = len(subexprs)
    new_n = merge_adjacent_default_or_grammars(subexprs, n)
    assert new_n == len(test_case["expected"]), "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Original: {n!s}",
            f"Expected: {len(test_case['expected'])!s}",
            f"Actual: {new_n!s}",
        )
    )
    assert subexprs == test_case["expected"], "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Original: {fmt_result(test_case['subexprs'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(subexprs)!s}",
        )
    )



import pytest

from grammatica.constants import (
    ALWAYS_SAFE_CHARS,
    CHAR_ESCAPE_MAP,
    STRING_LITERAL_ESCAPE_CHARS,
)
from grammatica.grammar import String, merge_adjacent_string_grammars
from grammatica.grammar.group import And

try:
    from .helpers import fmt_result
except ImportError:
    import sys
    from os import path

    sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))
    from helpers import fmt_result


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Empty",
            "grammar": String(""),
            "expected": None,
        },
        {
            "description": "Single character",
            "grammar": String("a"),
            "expected": '"a"',
        },
        {
            "description": "Multiple characters",
            "grammar": String("abc"),
            "expected": '"abc"',
        },
        {
            "description": "Always safe characters",
            "grammar": String("".join(ALWAYS_SAFE_CHARS)),
            "expected": '"{}"'.format("".join(ALWAYS_SAFE_CHARS)),
        },
        {
            "description": "Escape general tokens",
            "grammar": String("".join(CHAR_ESCAPE_MAP)),
            "expected": '"{}"'.format("".join(CHAR_ESCAPE_MAP.values())),
        },
        {
            "description": "Escape string literal characters",
            "grammar": String("".join(STRING_LITERAL_ESCAPE_CHARS)),
            "expected": '"{}"'.format("".join(f"\\{c}" for c in STRING_LITERAL_ESCAPE_CHARS)),
        },
        {
            "description": "Escape non-ASCII characters",
            "grammar": String("".join(chr(i) for i in range(127462, 127462 + 26))),
            "expected": '"{}"'.format("".join(f"\\x{i:02X}" for i in range(127462, 127462 + 26))),
        },
    ],
)
def test_string_render(test_case):
    actual = test_case["grammar"].render()
    assert actual == test_case["expected"], "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(actual)!s}",
        )
    )


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Empty",
            "grammar": String(""),
            "expected": None,
        },
        {
            "description": "Single character",
            "grammar": String("a"),
            "expected": String("a"),
        },
        {
            "description": "Multiple characters",
            "grammar": String("abc"),
            "expected": String("abc"),
        },
    ],
)
def test_string_simplify(test_case):
    actual = test_case["grammar"].simplify()
    assert (actual == test_case["expected"]) and (
        (test_case["expected"] is not actual) or (test_case["expected"] is actual is None)
    ), "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(actual)!s}",
        )
    )


def test_string_attrs_dict():
    string = String("test")
    assert string.attrs_dict() == {"value": "test"}


def test_string_as_string():
    string = String("test")
    assert string.as_string(indent=None) == string.as_string(indent=2) == "String(value='test')"


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Empty list",
            "subexprs": [],
            "expected": [],
        },
        {
            "description": "Single String",
            "subexprs": [String("a")],
            "expected": [String("a")],
        },
        {
            "description": "Multiple adjacent String",
            "subexprs": [String("a"), String("b"), String("c")],
            "expected": [String("abc")],
        },
        {
            "description": "Mixed adjacent String and other grammars",
            "subexprs": [String("a"), String("b"), And([String("c")]), String("d")],
            "expected": [String("ab"), And([String("c")]), String("d")],
        },
        {
            "description": "No adjacent String",
            "subexprs": [And([String("a")]), And([String("b")])],
            "expected": [And([String("a")]), And([String("b")])],
        },
    ],
)
def test_merge_adjacent_string_grammars(test_case):
    subexprs = [expr.copy() for expr in test_case["subexprs"]]
    n = len(subexprs)
    new_n = merge_adjacent_string_grammars(subexprs, n)
    assert new_n == len(test_case["expected"]), "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Original: {n!s}",
            f"Expected: {len(test_case['expected'])!s}",
            f"Actual: {new_n!s}",
        )
    )
    assert subexprs == test_case["expected"], "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Original: {fmt_result(test_case['subexprs'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(subexprs)!s}",
        )
    )



import pytest

from grammatica.grammar import String
from grammatica.grammar.base import value_is_simple, value_to_string

try:
    from .helpers import (
        NoOpGrammar,
        NoOpGrammarAlt,
        NoOpGroupGrammar,
        fmt_result,
    )
except ImportError:
    import sys
    from os import path

    sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))
    from helpers import (
        NoOpGrammar,
        NoOpGrammarAlt,
        NoOpGroupGrammar,
        fmt_result,
    )


def test_grammar_render():
    grammar = NoOpGrammar()
    assert super(NoOpGrammar, grammar).render() is None


def test_grammar_simplify():
    grammar = NoOpGrammar()
    assert super(NoOpGrammar, grammar).simplify() is None


def test_grammar_attrs_dict():
    grammar = NoOpGrammar()
    assert super(NoOpGrammar, grammar).attrs_dict() == {}


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Not equal to non-grammar type: None",
            "grammar": NoOpGrammar(),
            "other": None,
            "check_quantifier": True,
            "expected": False,
        },
        {
            "description": "Not equal to non-grammar type: str",
            "grammar": NoOpGrammar(),
            "other": "test",
            "check_quantifier": True,
            "expected": False,
        },
        {
            "description": "Not equal to non-grammar type: int",
            "grammar": NoOpGrammar(),
            "other": 42,
            "check_quantifier": True,
            "expected": False,
        },
        {
            "description": "Not equal to non-grammar type: float",
            "grammar": NoOpGrammar(),
            "other": 3.14,
            "check_quantifier": True,
            "expected": False,
        },
        {
            "description": "Equal to same type and same attributes",
            "grammar": NoOpGrammar(),
            "other": NoOpGrammar(),
            "check_quantifier": True,
            "expected": True,
        },
        {
            "description": "Equal to same type and attributes",
            "grammar": NoOpGroupGrammar([String("a")], quantifier=(1, 1)),
            "other": NoOpGroupGrammar([String("a")], quantifier=(1, 1)),
            "check_quantifier": True,
            "expected": True,
        },
        {
            "description": "Not equal to same type and different attributes",
            "grammar": NoOpGroupGrammar([String("a")], quantifier=(1, 1)),
            "other": NoOpGroupGrammar([String("b")], quantifier=(1, 1)),
            "check_quantifier": True,
            "expected": False,
        },
        {
            "description": "Not equal to different grammar type even with the same attributes",
            "grammar": NoOpGrammar(),
            "other": NoOpGrammarAlt(),
            "check_quantifier": True,
            "expected": False,
        },
        {
            "description": "Not equal when quantifier differs and check_quantifier is True",
            "grammar": NoOpGroupGrammar([String("a")], quantifier=(1, 1)),
            "other": NoOpGroupGrammar([String("a")], quantifier=(2, 2)),
            "check_quantifier": True,
            "expected": False,
        },
        {
            "description": "Equal when quantifier differs and check_quantifier is False",
            "grammar": NoOpGroupGrammar([String("a")], quantifier=(1, 1)),
            "other": NoOpGroupGrammar([String("a")], quantifier=(2, 2)),
            "check_quantifier": False,
            "expected": True,
        },
        {
            "description": "Different String instances",
            "grammar": String("test"),
            "other": String("different"),
            "check_quantifier": False,
            "expected": False,
        },
    ],
)
def test_grammar_equals(test_case):
    actual = test_case["grammar"].equals(
        test_case["other"],
        check_quantifier=test_case["check_quantifier"],
    )
    assert actual is test_case["expected"], "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Other: {fmt_result(test_case['other'])!s}",
            f"Expected: {test_case['expected']!s}",
            f"Actual: {actual!s}",
        )
    )


def test_grammar_equals_same_instance():
    grammar = NoOpGrammar()
    assert grammar.equals(grammar)


# TODO: More direct way to test would be to have a child of Grammar that defines some attrs
@pytest.mark.parametrize(
    "grammar, indent, expected",
    [
        (
            NoOpGroupGrammar([String("a")]),
            None,
            "NoOpGroupGrammar(subexprs=[String(value='a')], quantifier=(1, 1))",
        ),
        (
            NoOpGroupGrammar([String("a")]),
            2,
            "NoOpGroupGrammar(\n  subexprs=[\n    String(value='a')\n  ],\n  quantifier=(1, 1)\n)",
        ),
    ],
)
def test_grammar_as_string(grammar, indent, expected):
    assert grammar.as_string(indent=indent) == expected


def test_grammar_str_and_repr():
    grammar = NoOpGroupGrammar([String("a"), String("b")])
    assert (
        str(grammar)
        == repr(grammar)
        == grammar.as_string(indent=None)
        == "NoOpGroupGrammar(subexprs=[String(value='a'), String(value='b')], quantifier=(1, 1))"
    )


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, True),
        (42, True),
        (3.14, True),
        ("string", True),
        (True, True),
        (False, True),
        ([], False),
        (tuple(), False),
        ({}, False),
        (set(), False),
        (frozenset(), False),
        (String("test"), False),
    ],
)
def test_value_is_simple(value, expected):
    assert value_is_simple(value) == expected


@pytest.mark.parametrize(
    "value, indent, expected",
    [
        # None
        (None, None, "None"),
        # Boolean
        (False, None, "False"),
        (True, None, "True"),
        # Number
        (0, None, "0"),
        (42, None, "42"),
        (0.0, None, "0.0"),
        (3.14, None, "3.14"),
        # String
        ("", None, "''"),
        ("test", None, "'test'"),
        # List
        ([], None, "[]"),
        ([], 2, "[]"),
        ([1, 2], None, "[1, 2]"),
        ([1, 2], 2, "[1, 2]"),
        # Tuple
        (tuple(), None, "()"),
        (tuple(), 2, "()"),
        ((1, 2), None, "(1, 2)"),
        ((1, 2), 2, "(1, 2)"),
        # Set
        (set(), None, "set()"),
        (set(), 2, "set()"),
        ({1, 2}, None, "{1, 2}"),
        ({1, 2}, 2, "{1, 2}"),
        (frozenset(), None, "frozenset()"),
        (frozenset(), 2, "frozenset()"),
        (frozenset({1, 2}), None, "frozenset({1, 2})"),
        (frozenset({1, 2}), 2, "frozenset({1, 2})"),
        # Dict
        ({}, None, "{}"),
        ({}, 2, "{}"),
        ({"key": "value"}, None, "{'key': 'value'}"),
        # Grammar
        (String("a"), None, "String(value='a')"),
        ([String("a"), String("b")], None, "[String(value='a'), String(value='b')]"),
        ([String("a"), String("b")], 2, "[\n  String(value='a'),\n  String(value='b')\n]"),
        ((String("a"), String("b")), None, "(String(value='a'), String(value='b'))"),
        ((String("a"), String("b")), 2, "(\n  String(value='a'),\n  String(value='b')\n)"),
        (
            {"a": String("a"), "b": String("b")},
            None,
            "{'a': String(value='a'), 'b': String(value='b')}",
        ),
        (
            {"a": String("a"), "b": String("b")},
            2,
            "{\n  'a': String(value='a'),\n  'b': String(value='b')\n}",
        ),
    ],
)
def test_value_to_string(value, indent, expected):
    assert value_to_string(value, indent=indent) == expected


def test_value_to_string_unsupported_type():
    class CustomType:
        pass

    with pytest.raises(ValueError, match=r"Unsupported value type: CustomType"):
        value_to_string(CustomType(), indent=None)



"""
Classes and utilities for string literal expressions.
"""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, cast

from grammatica.constants import (
    ALWAYS_SAFE_CHARS,
    CHAR_ESCAPE_MAP,
    STRING_LITERAL_ESCAPE_CHARS,
)
from grammatica.grammar.base import Grammar
from grammatica.utils import char_to_hex

if sys.version_info >= (3, 12):  # pragma: no cover
    from typing import override
else:  # pragma: no cover
    from typing_extensions import override

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any


class String(Grammar):
    """Grammar that exactly matches a string.

    Args:
        value (Iterable[str]): Characters to match exactly.
    """

    __slots__: tuple[str, ...] = ("value",)

    def __init__(self, value: Iterable[str]) -> None:
        super().__init__()

        self.value: str = str(value)
        """String to match exactly."""

    def render(self, full: bool = True, wrap: bool = True) -> str | None:
        if len(self.value) == 0:
            return None
        return '"' + "".join(map(self._escape, self.value)) + '"'

    def simplify(self) -> String | None:
        if len(self.value) == 0:
            return None
        return String(self.value)

    @staticmethod
    def _escape(char: str) -> str:
        """Escape a character for use in a string literal.

        Args:
            char (str): Character to escape.

        Returns:
            str: Escaped character.
        """
        if char in ALWAYS_SAFE_CHARS:
            return char
        if char in CHAR_ESCAPE_MAP:
            return CHAR_ESCAPE_MAP[char]
        if char in STRING_LITERAL_ESCAPE_CHARS:
            return "\\" + char
        return char_to_hex(char)

    @override
    def as_string(self, indent: int | None = None) -> str:
        return super().as_string(indent=None)

    def attrs_dict(self) -> dict[str, Any]:
        return {"value": self.value}


def merge_adjacent_string_grammars(subexprs: list[Grammar], n: int) -> int:
    """Merge adjacent String grammars in-place. abc

    Args:
        subexprs (list[Grammar]): Subexpressions to merge.
        n (int): Number of subexpressions.

    Returns:
        int: Number of subexpressions after merging.
    """
    if n < 2:
        return n

    last_idx = -1
    for i in range(n - 1, -1, -1):
        if isinstance(subexprs[i], String):
            if last_idx < 1:
                last_idx = i
            else:
                n -= 1
        else:
            if last_idx > 0:
                subexprs[i + 1] = String(
                    "".join(
                        cast(String, s).value for s in subexprs[i + 1 : last_idx + 1]
                    )
                )
                del subexprs[i + 2 : last_idx + 1]
            last_idx = -1

    if last_idx > 0:
        subexprs[0] = String(
            "".join(cast(String, s).value for s in subexprs[0 : last_idx + 1])
        )
        del subexprs[1 : last_idx + 1]

    return n



"""
Utility functions.
"""

__all__ = [
    "char_to_cpoint",
    "char_to_hex",
    "ord_to_cpoint",
    "ord_to_hex",
]


def char_to_cpoint(char: str) -> str:
    return ord_to_cpoint(ord(char))


def ord_to_cpoint(ordinal: int) -> str:
    if ordinal < 0x10000:
        return f"\\u{ordinal:04X}"
    return f"\\U{ordinal:08X}"


def char_to_hex(char: str) -> str:
    return ord_to_hex(ord(char))


def ord_to_hex(ordinal: int) -> str:
    return f"\\x{ordinal:02X}"



import pytest

from grammatica.constants import (
    ALWAYS_SAFE_CHARS,
    CHAR_ESCAPE_MAP,
    RANGE_ESCAPE_CHARS,
)
from grammatica.grammar.char_range import CharRange
from grammatica.grammar.string import String

try:
    from .helpers import fmt_result
except ImportError:
    import sys
    from os import path

    sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))
    from helpers import fmt_result


def test_char_range_render_empty():
    grammar = CharRange([("a", "z")])
    grammar.char_ranges = []
    assert grammar.render() is None


def test_char_range_from_chars():
    assert CharRange.from_chars("abcxyz") == CharRange([("a", "c"), ("x", "z")])


def test_char_range_from_chars_negated():
    assert CharRange.from_chars("xyz", negate=True) == CharRange([("x", "z")], negate=True)


def test_char_range_from_ords():
    assert CharRange.from_ords([97, 98, 99, 120, 121, 122]) == CharRange([("a", "c"), ("x", "z")])


def test_char_range_from_ords_negated():
    assert CharRange.from_ords([97, 98, 99, 120, 121, 122], negate=True) == CharRange([("a", "c"), ("x", "z")], negate=True)


def test_char_range_validation_empty():
    with pytest.raises(ValueError, match="char_ranges must not be empty"):
        CharRange([])


def test_char_range_validation_start_length():
    with pytest.raises(ValueError, match="start must be a single character"):
        CharRange([("ab", "z")])


def test_char_range_validation_end_length():
    with pytest.raises(ValueError, match="end must be a single character"):
        CharRange([("a", "zz")])


def test_char_range_validation_end_before_start():
    with pytest.raises(ValueError, match="end must be greater than or equal to start"):
        CharRange([("z", "a")])


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Single character simplifies to String",
            "grammar": CharRange([("a", "a")]),
            "expected": String("a"),
        },
        {
            "description": "CharRange stays as CharRange",
            "grammar": CharRange([("a", "z")]),
            "expected": CharRange([("a", "z")]),
        },
        {
            "description": "Multiple ranges stay as CharRange",
            "grammar": CharRange([("a", "c"), ("x", "z")]),
            "expected": CharRange([("a", "c"), ("x", "z")]),
        },
        {
            "description": "Overlapping ranges merge",
            "grammar": CharRange([("a", "c"), ("b", "e")]),
            "expected": CharRange([("a", "e")]),
        },
    ],
)
def test_char_range_simplify(test_case):
    actual = test_case["grammar"].simplify()
    assert (actual == test_case["expected"]) and (
        (test_case["expected"] is not actual) or (test_case["expected"] is actual is None)
    ), "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(actual)!s}",
        )
    )


def test_char_range_simplify_empty():
    grammar = CharRange([("a", "z")])
    grammar.char_ranges = []
    assert grammar.simplify() is None


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Simple range",
            "grammar": CharRange([("a", "z")]),
            "expected": "[a-z]",
        },
        {
            "description": "Single character",
            "grammar": CharRange([("a", "a")]),
            "expected": "[a]",
        },
        {
            "description": "Multiple ranges",
            "grammar": CharRange([("a", "z"), ("0", "9")]),
            "expected": "[0-9a-z]",
        },
        {
            "description": "Negated range",
            "grammar": CharRange([("a", "z")], negate=True),
            "expected": "[^a-z]",
        },
        {
            "description": "Adjacent characters",
            "grammar": CharRange([("a", "b")]),
            "expected": "[ab]",
        },
    ]
    + [
        {
            "description": f"Character is always safe in ranges: {char!r}",
            "grammar": CharRange([(char, char)]),
            "expected": f"[{char}]",
        }
        for char in ALWAYS_SAFE_CHARS - frozenset(CHAR_ESCAPE_MAP) - RANGE_ESCAPE_CHARS
    ]
    + [
        {
            "description": f"Character is always escaped in ranges: {char!r}",
            "grammar": CharRange([(char, char)]),
            "expected": f"[{CHAR_ESCAPE_MAP[char]}]",
        }
        for char in frozenset(CHAR_ESCAPE_MAP) - RANGE_ESCAPE_CHARS
    ]
    + [
        {
            "description": f"Character is always simple-escaped in a ranges: {char!r}",
            "grammar": CharRange([(char, char)]),
            "expected": f"[\\{char}]",
        }
        for char in RANGE_ESCAPE_CHARS
    ]
    + [
        {
            "description": f"Other character is fully-escaped in ranges: {char!r}",
            "grammar": CharRange([(char, char)]),
            "expected": f"[\\x{ord(char):02X}]",
        }
        for char in map(chr, frozenset({0x00, 0x01, 0x02, 0x03, 0x04, 0x7F, 0x80, 0x81, 0xFF}))
    ],
)
def test_char_range_render(test_case):
    actual = test_case["grammar"].render()
    assert actual == test_case["expected"], "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!r}",
            f"Actual: {fmt_result(actual)!r}",
        )
    )


def test_char_range_as_string():
    grammar = CharRange([("a", "z")])
    assert grammar.as_string(indent=None) == grammar.as_string(indent=2) == "CharRange(char_ranges=[('a', 'z')], negate=False)"


def test_char_range_attrs_dict():
    grammar = CharRange([("a", "z")])
    assert grammar.attrs_dict() == {
        "char_ranges": [("a", "z")],
        "negate": False,
    }



from grammatica.grammar.group.and_ import (
    And,
    merge_adjacent_default_and_grammars,
    merge_adjacent_string_grammars,
)
from grammatica.grammar.group.base import GroupGrammar
from grammatica.grammar.group.or_ import Or, merge_adjacent_default_or_grammars

__all__ = [
    "GroupGrammar",
    "And",
    "merge_adjacent_default_and_grammars",
    "merge_adjacent_string_grammars",
    "Or",
    "merge_adjacent_default_or_grammars",
]



import pytest

from grammatica.utils import (
    char_to_cpoint,
    char_to_hex,
    ord_to_cpoint,
    ord_to_hex,
)


@pytest.mark.parametrize(
    "char,expected",
    [
        ("a", "\\u0061"),
        ("A", "\\u0041"),
        ("0", "\\u0030"),
        ("\n", "\\u000A"),
        ("€", "\\u20AC"),
        ("😀", "\\U0001F600"),  # Emoji (5+ char codepoint)
    ],
)
def test_char_to_cpoint(char, expected):
    assert char_to_cpoint(char) == expected


@pytest.mark.parametrize(
    "ordinal,expected",
    [
        (97, "\\u0061"),  # 'a'
        (65, "\\u0041"),  # 'A'
        (10, "\\u000A"),  # newline
        (8364, "\\u20AC"),  # '€'
        (128512, "\\U0001F600"),  # emoji (5+ digits)
    ],
)
def test_ord_to_cpoint(ordinal, expected):
    assert ord_to_cpoint(ordinal) == expected


@pytest.mark.parametrize(
    "char,expected",
    [
        ("a", "\\x61"),
        ("A", "\\x41"),
        ("0", "\\x30"),
        ("\n", "\\x0A"),
        ("~", "\\x7E"),
        ("😀", "\\x1F600"),
    ],
)
def test_char_to_hex(char, expected):
    assert char_to_hex(char) == expected


@pytest.mark.parametrize(
    "ordinal,expected",
    [
        (97, "\\x61"),  # 'a'
        (65, "\\x41"),  # 'A'
        (10, "\\x0A"),  # newline
        (126, "\\x7E"),  # '~'
        (255, "\\xFF"),  # max 2-digit hex
        (128512, "\\x1F600"),  # emoji (5+ digits)
    ],
)
def test_ord_to_hex(ordinal, expected):
    assert ord_to_hex(ordinal) == expected



import pytest

from grammatica.grammar import String

try:
    from ..helpers import NoOpGroupGrammar
except ImportError:
    import sys
    from os import path

    sys.path.append(path.dirname(path.dirname(path.dirname(path.abspath(__file__)))))
    from helpers import NoOpGroupGrammar


@pytest.mark.parametrize(
    "quantifier, expected",
    [
        (1, (1, 1)),
        (5, (5, 5)),
        ((1, 1), (1, 1)),
        ((0, 1), (0, 1)),
        ((0, 5), (0, 5)),
        ((0, None), (0, None)),
        ((1, None), (1, None)),
        ((5, 5), (5, 5)),
        ((2, 5), (2, 5)),
        ((2, None), (2, None)),
    ],
)
def test_group_grammar_quantifier(quantifier, expected):
    grammar = NoOpGroupGrammar([String("a")], quantifier=quantifier)
    assert grammar.quantifier == expected


def test_group_grammar_quantifier_validation_negative_lower_bound():
    with pytest.raises(ValueError, match=r"Range lower bound must be non-negative: \(-1, 5\)"):
        NoOpGroupGrammar([String("a")], quantifier=(-1, 5))


def test_group_grammar_quantifier_validation_zero_upper_bound():
    with pytest.raises(
        ValueError,
        match=r"Range upper bound must be positive or None \(infinity\): \(0, 0\)",
    ):
        NoOpGroupGrammar([String("a")], quantifier=(0, 0))


def test_group_grammar_quantifier_validation_negative_upper_bound():
    with pytest.raises(
        ValueError,
        match=r"Range upper bound must be positive or None \(infinity\): \(0, -1\)",
    ):
        NoOpGroupGrammar([String("a")], quantifier=(0, -1))


def test_group_grammar_quantifier_validation_lower_bound_greater_than_upper_bound():
    with pytest.raises(ValueError, match=r"Range lower bound must be <= range upper bound: \(5, 3\)"):
        NoOpGroupGrammar([], quantifier=(5, 3))


@pytest.mark.parametrize(
    "grammar, expected",
    [
        (NoOpGroupGrammar([]), None),
        (NoOpGroupGrammar([String("")]), None),
        (NoOpGroupGrammar([String("a"), String("")]), '"a"'),
    ],
)
def test_group_grammar_render(grammar, expected):
    actual = super(NoOpGroupGrammar, grammar).render()
    if expected is None:
        assert actual is None
    else:
        assert actual == expected


def test_group_grammar_simplify():
    grammar = NoOpGroupGrammar([])
    assert super(NoOpGroupGrammar, grammar).simplify() is None


def test_group_grammar_simplify_subexprs():
    grammar = NoOpGroupGrammar([])
    assert super(NoOpGroupGrammar, grammar).simplify_subexprs(grammar.subexprs, grammar.quantifier) is None


def test_group_grammar_needs_wrapped():
    grammar = NoOpGroupGrammar([])
    assert super(NoOpGroupGrammar, grammar).needs_wrapped() is False


def test_group_grammar_attrs_dict():
    grammar = NoOpGroupGrammar([])
    assert super(NoOpGroupGrammar, grammar).attrs_dict() == {
        "subexprs": [],
        "quantifier": (1, 1),
    }


@pytest.mark.parametrize(
    "quantifier, expected",
    [
        ((1, 1), None),
        ((0, 1), "?"),
        ((0, 2), "{0,2}"),
        ((0, None), "*"),
        ((1, None), "+"),
        ((2, 2), "{2}"),
        ((2, 5), "{2,5}"),
        ((2, None), "{2,}"),
    ],
)
def test_group_grammar_render_quantifier(quantifier, expected):
    grammar = NoOpGroupGrammar([], quantifier=quantifier)
    assert grammar.render_quantifier() == expected



"""
Classes and utilities for defining derivation rules.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from grammatica.grammar.base import Grammar

if TYPE_CHECKING:
    from typing import Any


class DerivationRule(Grammar):
    """Derivation rule that expands to a expression (grammar).

    Non-terminal symbol and an expression separated by a metasymbol (e.g. 'symbol ::= expression').

    Args:
        symbol (str): Symbol (non-terminal) for the derivation rule.
        value (Grammar): Grammar the symbol derives into.
    """

    __slots__: tuple[str, ...] = ("symbol", "value")

    separator: str = " ::= "
    """Separator metasymbol to use for the derivation rule."""

    def __init__(self, symbol: str, value: Grammar) -> None:
        super().__init__()

        self.symbol: str = symbol
        """Symbol (non-terminal) for the derivation rule."""
        self.value: Grammar = value
        """Grammar the symbol derives into."""

    def render(self, full: bool = True, wrap: bool = True) -> str | None:
        if not full:
            return self.symbol
        rendered = self.value.render(full=False, wrap=wrap)
        if not rendered:
            return None
        expr = f"{self.symbol}{self.separator}{rendered}"
        return expr

    def simplify(self) -> DerivationRule | None:
        simplified = self.value.simplify()
        if simplified is None:
            return None
        return DerivationRule(self.symbol, simplified)

    def attrs_dict(self) -> dict[str, Any]:
        return {"symbol": self.symbol, "value": self.value}



#!/usr/bin/env python3

import argparse
import http.server
import logging
import os
from dataclasses import dataclass
from pathlib import Path

PROJECT_DIR: Path = Path(__file__).parent.parent

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class Args:
    html_dir: Path
    host: str
    port: int
    verbosity: int


def parse_args() -> Args:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--html",
        type=Path,
        default=PROJECT_DIR / "docs" / "build" / "html",
        metavar="HTML_DIR",
        help="Directory containing HTML files to serve.",
    )
    parser.add_argument(
        "--host",
        type=str,
        default="localhost",
        metavar="HOST",
        help="Web server host to listen on.",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=9000,
        metavar="PORT",
        help="Web server port to listen on.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="verbosity",
        default=0,
        help="Increase logging verbosity (0-1).",
    )
    args = parser.parse_args()
    return Args(
        html_dir=args.html.resolve(),
        host=args.host,
        port=args.port,
        verbosity=args.verbosity,
    )


def main() -> None:
    args = parse_args()

    log_level: int
    if args.verbosity == 0:
        log_level = logging.INFO
    elif args.verbosity == 1:
        log_level = logging.DEBUG
    else:
        raise ValueError(f"Invalid verbosity level: {args.verbosity}")
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    if not args.html_dir.is_dir():
        raise FileNotFoundError(f"HTML directory not found: {args.html_dir}")

    os.chdir(args.html_dir)
    httpd = http.server.HTTPServer(
        (args.host, args.port),
        http.server.SimpleHTTPRequestHandler,
    )
    url = "http://" + ":".join(map(str, httpd.server_address))
    logger.info("Starting HTTP server on %s", url)
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        logger.info("Shutting down HTTP server")
        httpd.shutdown()
    finally:
        logger.info("Closing HTTP server")
        httpd.server_close()


if __name__ == "__main__":
    main()



from __future__ import annotations

import inspect
import os
import re
import sys
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import quote

from sphinx.builders.html import StandaloneHTMLBuilder

if TYPE_CHECKING:
    from typing import Any

    from sphinx.application import Sphinx


if sys.version_info >= (3, 12):  # pragma: no cover
    from typing import override
else:  # pragma: no cover
    from typing_extensions import override

sys.path.insert(0, str(Path(__file__).parent.parent.parent / "src"))

import grammatica  # pylint: disable=C0413

RUNNING_CI: bool = os.environ.get("RUNNING_CI", "0") == "1"
URL_PATH_PREFIX: str = os.environ.get("URL_PATH_PREFIX", "")

URI_SUFFIX: str = "" if RUNNING_CI else "index.html"
REPO_URL: str = "https://github.com/yaphott/grammatica"


class CustomHTMLBuilder(StandaloneHTMLBuilder):
    name = "html"

    @override
    def get_target_uri(self, docname: str, typ: str | None = None) -> str:
        if docname == "index":
            return self.config["html_baseurl"] + URI_SUFFIX
        if docname.endswith("/index"):
            docname = docname.removesuffix("/index")
        if docname:
            docname = quote(docname)
        return self.config["html_baseurl"] + docname + "/" + URI_SUFFIX

    @override
    def get_output_path(self, page_name: str, /) -> Path:
        if page_name == "index":
            return Path(self.outdir, "index.html")
        if page_name.endswith("/index"):
            page_name = page_name.removesuffix("/index")
        return Path(self.outdir, page_name, "index.html")


def setup(app: Sphinx) -> dict[str, Any]:
    app.setup_extension("sphinx.builders.html")
    app.add_builder(CustomHTMLBuilder, override=True)
    return {
        "version": grammatica.__version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


def linkcode_resolve(domain, info):
    if domain != "py":
        return None

    submodule = sys.modules.get(info["module"])
    if submodule is None:
        return None

    obj = submodule
    for part in info["fullname"].split("."):
        try:
            obj = getattr(obj, part)
        except AttributeError:
            return None

    try:
        func_path = inspect.getsourcefile(obj)
    except TypeError:
        return None
    if func_path is None:
        return None
    func_path = Path(func_path).relative_to(Path(grammatica.__file__).parent)

    # Strip decorators
    obj = inspect.unwrap(obj)

    # Ignore re-exports
    module = inspect.getmodule(obj)
    if (module is not None) and not module.__name__.startswith(f"{grammatica.__name__}."):
        return None

    if ".dev" in grammatica.__version__:
        ref = "main"
    else:
        ref = "v" + ".".join(grammatica.__version__.split(".")[:3])

    source, line_n = inspect.getsourcelines(obj)
    if line_n:
        line_spec = f"#L{line_n}-L{line_n + len(source) - 1}"
    else:
        line_spec = ""

    return "/".join(
        (
            REPO_URL,
            "blob",
            ref,
            "src",
            "grammatica",
            func_path.as_posix() + line_spec,
        ),
    )


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information ----------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

start_year = 2025
end_year = datetime.now(tz=timezone.utc).year

project = "Grammatica"
author = "Nicholas Londowski"
copyright = f"{start_year}-{end_year}, {author}" if end_year > start_year else f"{start_year}, {author}"
version = re.sub(r"(\d+\.\d+)\.\d+(?:(\.dev\d+).*)?", r"\1\2", grammatica.__version__)
release = grammatica.__version__

# -- General configuration ----------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

needs_sphinx = "8.2.3"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
]

source_encoding = "utf-8"
master_doc = "index"
templates_path = ["_templates"]
exclude_patterns = ["build", "**/.git", "Thumbs.db", ".DS_Store"]
# nitpicky = True

# -- Options for HTML output ----------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_baseurl = "/" + URL_PATH_PREFIX
html_static_path = ["_static"]
html_css_files = ["grammatica.css"]
html_js_files = [
    ("custom-icons.js", {"defer": "defer"}),
]
html_theme_options = {
    "logo": {
        "image_light": "_static/logo.svg",
        "image_dark": "_static/logo-dark.svg",
    },
    "collapse_navigation": True,
    "external_links": [
        {"name": "Londowski", "url": "https://londowski.com/"},
    ],
    "header_links_before_dropdown": 4,
    "navbar_align": "left",
    "navbar_start": ["navbar-logo"],
    "navbar_center": ["version-switcher", "navbar-nav"],
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
    "navbar_persistent": ["search-field"],
    "footer_start": ["copyright"],
    "footer_center": ["sphinx-version"],
    "switcher": {
        "version_match": version,
        "json_url": "https://londowski.com/docs/grammatica/versions.json",
    },
    "show_version_warning_banner": False,
    "icon_links": [
        {
            "name": "GitHub",
            "url": REPO_URL,
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/grammatica/",
            "icon": "fa-custom fa-pypi",
        },
        {
            "name": "Londowski",
            "url": "https://londowski.com/",
            "icon": "fa-solid fa-l",
        },
    ],
}
html_title = f"{project} v{grammatica.__version__} Manual"
html_last_updated_fmt = "%b %d, %Y"
html_context = {"default_mode": "light"}
html_favicon = "_static/favicon.ico"
html_copy_source = False

# -- Options for Autodoc ---------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

autodoc_member_order = "bysource"
autodoc_typehints_description_target = "documented"
autodoc_use_type_comments = False

# -- Options for sphinx-autodoc-typehints --------------------------------------
# https://github.com/tox-dev/sphinx-autodoc-typehints?tab=readme-ov-file#options

typehints_use_rtype = False
typehints_document_rtype_none = False

# -- Options for Autosummary -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autosummary.html#configuration

autosummary_generate = True

# -- Options for Napoleon settings ------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html#configuration

napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = True
always_use_bars_union = True
napoleon_use_ivar = False
napoleon_use_rtype = False
napoleon_use_param = False
napoleon_preprocess_types = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True



from __future__ import annotations

from typing import TYPE_CHECKING

from grammatica.grammar.base import Grammar, value_to_string
from grammatica.grammar.group.base import GroupGrammar

if TYPE_CHECKING:
    from typing import Any


class NoOpGrammar(Grammar):
    def __init__(self):
        pass

    def render(self, wrap=False):
        return self.render()

    def simplify(self):
        return self.simplify()

    def attrs_dict(self):
        return {}


class NoOpGrammarAlt(Grammar):
    def __init__(self):
        pass

    def render(self, wrap=False):
        return self.render()

    def simplify(self):
        return self.simplify()

    def attrs_dict(self):
        return {}


class NoOpGroupGrammar(GroupGrammar):
    separator = " "

    def __init__(
        self,
        exprs,
        quantifier=None,
    ):
        super().__init__(
            subexprs=exprs,
            quantifier=quantifier if quantifier else (1, 1),
        )

    @staticmethod
    def simplify_subexprs(original_subexprs, quantifier):
        return None

    def needs_wrapped(self):
        return False

    def attrs_dict(self):
        return super().attrs_dict()


def fmt_result(result: Any, indent: int = 2) -> str:
    try:
        return value_to_string(result, indent=indent)
    except ValueError:
        return repr(result)



"""
Classes and utilities for logical OR operations.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from grammatica.grammar.group.and_ import And
from grammatica.grammar.group.base import GroupGrammar

if TYPE_CHECKING:
    from collections.abc import Iterable

    from grammatica.grammar.base import Grammar


class Or(GroupGrammar):
    """Grouped grammar that represents a logical OR operation between grammars.

    Args:
        subexprs (Iterable[Grammar]): Grammars to group with OR logic.
        quantifier (int | tuple[int, int | None]): Minimum and maximum repetitions the expression must match. Defaults to (1, 1).

    Raises:
        ValueError: Range lower bound is negative.
        ValueError: Range upper bound is not positive or None (infinity).
        ValueError: Range lower bound is greater than range upper bound.
    """

    separator: str = " | "
    """Separator to use for the grammar."""

    def __init__(
        self,
        subexprs: Iterable[Grammar],
        quantifier: int | tuple[int, int | None] = (1, 1),
    ) -> None:
        super().__init__(
            subexprs=subexprs,
            quantifier=quantifier,
        )

    def needs_wrapped(self) -> bool:
        n = len(self.subexprs)
        if n < 1:
            return False
        if n == 1:
            if self.quantifier == (1, 1):
                return False
            # Recursively look for the first non-default (1, 1) subexpression
            subexpr = self.subexprs[0]
            wrap = isinstance(subexpr, GroupGrammar)
            while (
                wrap
                and (cast(GroupGrammar, subexpr).quantifier == (1, 1))
                and (len(cast(GroupGrammar, subexpr).subexprs) == 1)
            ):
                subexpr = cast(GroupGrammar, subexpr).subexprs[0]
                wrap = isinstance(subexpr, GroupGrammar)
            return wrap
        return True

    @staticmethod
    def simplify_subexprs(
        original_subexprs: list[Grammar],
        quantifier: tuple[int, int | None],
    ) -> Grammar | None:
        subexprs: list[Grammar] = []
        n = 0
        for subexpr in original_subexprs:
            simplified = subexpr.simplify()
            if simplified is None:
                continue
            if simplified in subexprs:
                continue
            subexprs.append(simplified)
            n += 1

        # Empty expression is no-op
        if n < 1:
            return None

        if n == 1:
            # Unwrap a single default (1, 1) subexpression
            if quantifier == (1, 1):
                return subexprs[0]
            # Or that is optional (0, 1) can recursively unwrap to the first simple or non-single, non-default (1, 1), and non-optional (0, 1) grouped grammar.
            if (
                (quantifier == (0, 1))
                and isinstance(subexprs[0], GroupGrammar)
                and (subexprs[0].quantifier in ((0, 1), (1, 1)))
            ):
                return And.simplify_subexprs(subexprs[0].subexprs, quantifier)
            # Or with a single subexpression is the same as And with a single subexpression
            return And.simplify_subexprs(subexprs, quantifier)

        merged_n = merge_adjacent_default_or_grammars(subexprs, n)
        if merged_n < n:
            return Or.simplify_subexprs(subexprs, quantifier)

        # Or that is optional (0, n) containing subexpressions that are all optional (0, n) does not need to then be optional itself
        if (quantifier[0] == 0) and all(
            map(
                lambda x: isinstance(x, GroupGrammar) and (x.quantifier[0] == 0),
                subexprs,
            )
        ):
            return Or.simplify_subexprs(subexprs, (1, quantifier[1]))

        return Or(subexprs, quantifier=quantifier)


def merge_adjacent_default_or_grammars(subexprs: list[Grammar], n: int) -> int:
    """Merge adjacent Or expressions having default quantifier (1, 1) in-place.

    Args:
        subexprs (list[Grammar]): Subexpressions to merge.
        n (int): Number of subexpressions.

    Returns:
        int: Number of subexpressions after merging.
    """
    if n < 2:
        return n
    last_idx = -1
    for i in range(n - 1, -1, -1):
        if isinstance(subexprs[i], Or) and (cast(Or, subexprs[i]).quantifier == (1, 1)):
            if last_idx < 1:
                last_idx = i
            else:
                n -= 1
        else:
            if last_idx > 0:
                subexprs[i + 1] = Or(
                    (
                        s
                        for subexpr in subexprs[i + 1 : last_idx + 1]
                        for s in cast(Or, subexpr).subexprs
                    )
                )
                del subexprs[i + 2 : last_idx + 1]
            last_idx = -1
    if last_idx > 0:
        subexprs[0] = Or(
            (
                s
                for subexpr in subexprs[0 : last_idx + 1]
                for s in cast(Or, subexpr).subexprs
            )
        )
        del subexprs[1 : last_idx + 1]
    return n



import pytest

from grammatica.constants import (
    ALWAYS_SAFE_CHARS,
    CHAR_ESCAPE_MAP,
    RANGE_ESCAPE_CHARS,
    STRING_LITERAL_ESCAPE_CHARS,
)


@pytest.mark.parametrize(
    "char, expected",
    [
        # Digits
        ("0", True),
        ("5", True),
        ("9", True),
        # Lowercase letters
        ("a", True),
        ("m", True),
        ("z", True),
        # Uppercase letters
        ("A", True),
        ("M", True),
        ("Z", True),
        # Space
        (" ", True),
        # Punctuation
        ("!", True),
        ("#", True),
        ("$", True),
        ("%", True),
        ("&", True),
        ("'", True),
        ("(", True),
        (")", True),
        ("*", True),
        ("+", True),
        (",", True),
        ("-", True),
        (".", True),
        ("/", True),
        (":", True),
        (";", True),
        ("<", True),
        ("=", True),
        (">", True),
        ("?", True),
        ("@", True),
        ("[", True),
        ("]", True),
        ("^", True),
        ("_", True),
        ("`", True),
        ("{", True),
        ("|", True),
        ("}", True),
        ("~", True),
        # Not safe
        ("\n", False),
        ("\r", False),
        ("\t", False),
        ('"', False),
        ("\\", False),
        ("\x80", False),
        ("€", False),
        ("😀", False),
    ],
)
def test_always_safe_chars(char, expected):
    assert (char in ALWAYS_SAFE_CHARS) == expected


@pytest.mark.parametrize(
    "char, expected",
    [
        ("\n", "\\n"),
        ("\r", "\\r"),
        ("\t", "\\t"),
    ],
)
def test_char_escape_map_has_escape(char, expected):
    assert char in CHAR_ESCAPE_MAP
    assert CHAR_ESCAPE_MAP[char] == expected


@pytest.mark.parametrize(
    "char",
    [
        "a",
        " ",
        "0",
        '"',
        "\\",
        "^",
        "-",
        "[",
        "]",
    ],
)
def test_char_escape_map_no_escape(char):
    assert char not in CHAR_ESCAPE_MAP


@pytest.mark.parametrize(
    "char, expected",
    [
        ('"', True),
        ("\\", True),
        ("a", False),
        ("[", False),
        ("-", False),
        ("^", False),
        ("]", False),
        (" ", False),
    ],
)
def test_string_literal_escape_chars(char, expected):
    assert (char in STRING_LITERAL_ESCAPE_CHARS) == expected


@pytest.mark.parametrize(
    "char, expected",
    [
        ("^", True),
        ("-", True),
        ("[", True),
        ("]", True),
        ("\\", True),
        ("a", False),
        ('"', False),
        (" ", False),
    ],
)
def test_range_escape_chars(char, expected):
    assert (char in RANGE_ESCAPE_CHARS) == expected



__all__ = [
    "ALWAYS_SAFE_CHARS",
    "CHAR_ESCAPE_MAP",
    "RANGE_ESCAPE_CHARS",
    "STRING_LITERAL_ESCAPE_CHARS",
]

_DIGITS: frozenset[str] = frozenset(map(chr, b"0123456789"))
_ASCIILETTERS: frozenset[str] = frozenset(
    map(chr, b"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")
)
_PUNCTUATION: frozenset[str] = frozenset(map(chr, b"!#$%&'()*+,-./:;<=>?@[]^_`{|}~"))
_SPACE: frozenset[str] = frozenset(map(chr, b" "))
ALWAYS_SAFE_CHARS: frozenset[str] = _DIGITS | _ASCIILETTERS | _PUNCTUATION | _SPACE
CHAR_ESCAPE_MAP: dict[str, str] = {
    "\n": "\\n",
    "\r": "\\r",
    "\t": "\\t",
}
STRING_LITERAL_ESCAPE_CHARS: frozenset[str] = frozenset(map(chr, b'"\\'))
RANGE_ESCAPE_CHARS: frozenset[str] = frozenset(map(chr, b"^-[]\\"))



"""Base grammar classes and utility functions for handling grammar expressions.

Provides abstractions for building and rendering grammar expressions, as well as utility functions for
string conversion and determining the simplicity of values.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import (
        Any,
        Collection,
        Mapping,
    )


class Grammar(ABC):
    """Base class for grammar expressions."""

    __slots__: tuple[str, ...] = ()

    @abstractmethod
    def render(self, full: bool = True, wrap: bool = True) -> str | None:
        """Render the grammar as a regular expression.

        Args:
            full (bool, optional): Render the full expression. Defaults to True.
            wrap (bool, optional): Wrap the expression in parentheses. Defaults to True.

        Returns:
            str | None: Rendered expression, or None if resolved to empty.
        """
        return None

    @abstractmethod
    def simplify(self) -> Grammar | None:
        """Simplify the grammar.

        Attempts to reduce redundancy, remove empty subexpressions, and optimize the grammar.

        Note:
            The resulting grammar and its parts are copies, and the original grammar is not modified.

        Returns:
            Grammar | None: Simplified expression, or None if resolved to empty.
        """
        return None

    @abstractmethod
    def attrs_dict(self) -> dict[str, Any]:
        """Return instance attributes of the grammar.

        Returns:
            dict[str, Any]: Dictionary of instance attributes.
        """
        return {}

    def copy(self) -> Grammar:
        """Create a copy of the grammar.

        Returns:
            Grammar: Copy of the grammar.
        """
        cls = type(self)
        kwargs = self.attrs_dict()
        g = cls(**kwargs)
        return g

    def as_string(self, indent: int | None = None) -> str:
        """Return a string representation of the grammar.

        Args:
            indent (int, optional): Number of spaces to indent each level. Defaults to None.

        Returns:
            str: String representation of the grammar.

        Raises:
            ValueError: Attribute type is not supported.
        """
        attrs = self.attrs_dict()
        n = len(attrs)
        msg = f"{type(self).__name__}("
        for j, (name, value) in enumerate(attrs.items()):
            if indent is None:
                if j > 0:
                    msg += ", "
            else:
                if j > 0:
                    msg += ","
                msg += "\n" + (" " * indent)
            msg += f"{name}="
            if indent is None:
                msg += value_to_string(value, indent=indent)
            else:
                msg += value_to_string(value, indent=indent).replace(
                    "\n",
                    "\n" + (" " * indent),
                )
                if j == n - 1:
                    msg += "\n"
        msg += ")"
        return msg

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return self.as_string(indent=None)

    def equals(self, other: Any, check_quantifier: bool = True) -> bool:
        """Check equality with another value.

        Args:
            other (Any): Value to compare against.
            check_quantifier (bool, optional): Include the quantifier in the comparison. Defaults to True.

        Returns:
            bool: True if the values are equal, False otherwise.
        """
        if self is other:
            return True
        if not isinstance(other, type(self)):
            return False
        attrs = self.attrs_dict()
        other_attrs = other.attrs_dict()
        if not check_quantifier:
            return {k: attrs[k] for k in sorted(attrs) if (k != "quantifier")} == {
                k: other_attrs[k] for k in sorted(other_attrs) if (k != "quantifier")
            }
        return {k: attrs[k] for k in sorted(attrs)} == {
            k: other_attrs[k] for k in sorted(other_attrs)
        }

    def __eq__(self, other: Any) -> bool:
        return self.equals(other)


def value_is_simple(value: Any) -> bool:
    """Determine if a value is simple (None, bool, int, float, or str).

    Args:
        value (Any): Value to check.

    Returns:
        bool: True if the value is simple, False otherwise.
    """
    if value is None:
        return True
    if isinstance(value, (int, float)):
        return True
    if isinstance(value, str):
        return True
    return False


def _collection_to_string(value: Collection[Any], indent: int | None) -> str:
    prefix: str
    suffix: str
    if isinstance(value, tuple):
        prefix, suffix = "(", ")"
    elif isinstance(value, list):
        prefix, suffix = "[", "]"
    elif isinstance(value, set):
        if not value:
            return "set()"
        prefix, suffix = "{", "}"
    else:  # frozenset
        if not value:
            return "frozenset()"
        prefix, suffix = "frozenset({", "})"
    msg = prefix
    if all(map(value_is_simple, value)):
        for j, subvalue in enumerate(value):
            if j > 0:
                msg += ", "
            msg += value_to_string(subvalue, indent=None)
    else:
        value_n = len(value)
        for j, subvalue in enumerate(value):
            if indent is None:
                if j > 0:
                    msg += ", "
                msg += value_to_string(subvalue, indent=indent)
            else:
                if j > 0:
                    msg += ","
                msg += "\n" + (" " * indent)
                msg += value_to_string(subvalue, indent=indent).replace(
                    "\n",
                    "\n" + (" " * indent),
                )
                if j == value_n - 1:
                    msg += "\n"
    msg += suffix
    return msg


def _mapping_to_string(value: Mapping[Any, Any], indent: int | None) -> str:
    msg = "{"
    for j, k in enumerate(tuple(value)):
        subkey, subvalue = k, value[k]
        if indent is None:
            if j > 0:
                msg += ", "
        else:
            if j > 0:
                msg += ","
            msg += "\n" + (" " * indent)
        msg += value_to_string(subkey, indent=indent) + ": "
        if indent is None:
            msg += value_to_string(subvalue, indent=indent)
        else:
            msg += value_to_string(subvalue, indent=indent).replace(
                "\n",
                "\n" + (" " * indent),
            )
            if j == len(value) - 1:
                msg += "\n"
    msg += "}"
    return msg


def value_to_string(value: Any, indent: int | None) -> str:
    """Create a string representation of a value.

    Args:
        value (Any): Value to represent as a string.
        indent (int | None): Indentation level, or None for no indentation.

    Returns:
        str: String representation of the value.

    Raises:
        ValueError: Value type is unsupported.
    """
    if value is None:
        return "None"
    if isinstance(value, bool):
        return "True" if value else "False"
    if isinstance(value, (int, float)):
        return str(value)
    if isinstance(value, str):
        return repr(value)
    if isinstance(value, (tuple, list, set, frozenset)):
        return _collection_to_string(value, indent=indent)
    if isinstance(value, dict):
        return _mapping_to_string(value, indent=indent)
    if isinstance(value, Grammar):
        return value.as_string(indent=indent)
    raise ValueError(f"Unsupported value type: {type(value).__name__}")



"""
Grammar classes and utility functions for handling grammar expressions.
"""

from grammatica.grammar.base import Grammar
from grammatica.grammar.char_range import CharRange
from grammatica.grammar.derivation_rule import DerivationRule
from grammatica.grammar.string import String, merge_adjacent_string_grammars

__all__ = [
    "Grammar",
    "CharRange",
    "DerivationRule",
    "String",
    "merge_adjacent_string_grammars",
]



#!/usr/bin/env python3

import sys


REQUIRED_PYTHON = (3, 10)
CURRENT_PYTHON = sys.version_info[:2]

if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write(
        """
==========================
Unsupported Python version
==========================
This version of Grammatica requires at least Python {}.{}, but you're trying to install it on Python {}.{}.
""".format(
            *(REQUIRED_PYTHON + CURRENT_PYTHON)
        )
    )
    sys.exit(1)

from setuptools import setup

setup()



"""
Grammatica is a GBNF (GGML BNF) grammar builder that exposes the full richness of the GBNF notation system using a syntax that reflects the resulting grammar.
"""

from importlib.metadata import PackageNotFoundError, version

__all__ = [
    "__version__",
]

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.0.dev0"



import pytest

from grammatica.grammar import DerivationRule, String
from grammatica.grammar.group import And, Or

try:
    from .helpers import fmt_result
except ImportError:
    import sys
    from os import path

    sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))
    from helpers import fmt_result


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Simple derivation rule",
            "grammar": DerivationRule("symbol", String("value")),
            "expected": 'symbol ::= "value"',
        },
        {
            "description": "Empty value returns None",
            "grammar": DerivationRule("symbol", String("")),
            "expected": None,
        },
        {
            "description": "With Grammar",
            "grammar": DerivationRule("expr", And([String("a"), String("b")])),
            "expected": 'expr ::= "a" "b"',
        },
        {
            "description": "With Or",
            "grammar": DerivationRule("choice", Or([String("a"), String("b")])),
            "expected": 'choice ::= ("a" | "b")',
        },
    ],
)
def test_derivation_rule_render_full(test_case):
    actual = test_case["grammar"].render(full=True)
    assert actual == test_case["expected"], "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(actual)!s}",
        )
    )


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Returns symbol when full=False",
            "grammar": DerivationRule("symbol", String("value")),
            "expected": "symbol",
        },
        {
            "description": "Returns symbol even with empty value",
            "grammar": DerivationRule("empty", String("")),
            "expected": "empty",
        },
    ],
)
def test_derivation_rule_render_not_full(test_case):
    actual = test_case["grammar"].render(full=False)
    assert actual == test_case["expected"], "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(actual)!s}",
        )
    )


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Wrap Or with multiple subexpressions when wrap=True",
            "grammar": DerivationRule("choice", Or([String("a"), String("b")])),
            "wrap": True,
            "expected": 'choice ::= ("a" | "b")',
        },
        {
            "description": "Do not wrap Or with single subexpression when wrap=True",
            "grammar": DerivationRule("choice", Or([String("a")])),
            "wrap": True,
            "expected": 'choice ::= "a"',
        },
        {
            "description": "Do not wrap Or with multiple subexpressions when wrap=False",
            "grammar": DerivationRule("choice", Or([String("a"), String("b")])),
            "wrap": False,
            "expected": 'choice ::= "a" | "b"',
        },
        {
            "description": "Do not wrap Or with single subexpression when wrap=False",
            "grammar": DerivationRule("choice", Or([String("a")])),
            "wrap": False,
            "expected": 'choice ::= "a"',
        },
        {
            "description": "Do not wrap And when wrap=True",
            "grammar": DerivationRule("choice", And([String("a"), String("b")])),
            "wrap": True,
            "expected": 'choice ::= "a" "b"',
        },
        {
            "description": "Do not wrap when wrap=False",
            "grammar": DerivationRule("choice", Or([String("a"), String("b")])),
            "wrap": False,
            "expected": 'choice ::= "a" | "b"',
        },
    ],
)
def test_derivation_rule_render_wrap(test_case):
    actual = test_case["grammar"].render(full=True, wrap=test_case["wrap"])
    assert actual == test_case["expected"], "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(actual)!s}",
        )
    )


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "description": "Simplifies value",
            "grammar": DerivationRule("merged", And([String("a"), String("b")])),
            "expected": DerivationRule("merged", String("ab")),
        },
        {
            "description": "Grammar that simplifies to None returns None",
            "grammar": DerivationRule("empty", And([])),
            "expected": None,
        },
    ],
)
def test_derivation_rule_simplify(test_case):
    actual = test_case["grammar"].simplify()
    assert (actual == test_case["expected"]) and (
        (test_case["expected"] is not actual) or (test_case["expected"] is actual is None)
    ), "\n".join(
        (
            f"Description: {test_case['description']!r}",
            f"Grammar: {fmt_result(test_case['grammar'])!s}",
            f"Expected: {fmt_result(test_case['expected'])!s}",
            f"Actual: {fmt_result(actual)!s}",
        )
    )


def test_derivation_rule_attrs_dict():
    symbol = "test_symbol"
    grammar = String("test_value")
    rule = DerivationRule(symbol, grammar)
    actual = rule.attrs_dict()
    expected = {
        "symbol": symbol,
        "value": grammar,
    }
    assert actual == expected
    assert actual["value"] is grammar



"""
Classes and utilities for logical AND operations.
"""

from __future__ import annotations

from collections import namedtuple
from typing import TYPE_CHECKING, cast

from grammatica.grammar.group.base import GroupGrammar
from grammatica.grammar.string import merge_adjacent_string_grammars

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from grammatica.grammar.base import Grammar


GroupWeight: type[tuple] = namedtuple(
    "GroupWeight",
    ["removed_size", "chunk_size", "count"],
)


class And(GroupGrammar):
    """Grouped grammar that represents a logical AND operation between grammars.

    Args:
        subexprs (Iterable[Grammar]): Grammars to group with AND logic.
        quantifier (int | tuple[int, int | None]): Minimum and maximum repetitions the expression must match. Defaults to (1, 1).

    Raises:
        ValueError: Range lower bound is negative.
        ValueError: Range upper bound is not positive or None (infinity).
        ValueError: Range lower bound is greater than range upper bound.
    """

    separator: str = " "
    """Separator to use for the grammar."""

    def __init__(
        self,
        subexprs: Iterable[Grammar],
        quantifier: int | tuple[int, int | None] = (1, 1),
    ) -> None:
        super().__init__(
            subexprs=subexprs,
            quantifier=quantifier,
        )

    def needs_wrapped(self) -> bool:
        n = len(self.subexprs)
        if n < 1:
            return False
        if n == 1:
            if self.quantifier == (1, 1):
                return False
            # Recursively look for the first non-default (1, 1) subexpression
            subexpr = self.subexprs[0]
            wrap = isinstance(subexpr, GroupGrammar)
            while (
                wrap
                and (cast(GroupGrammar, subexpr).quantifier == (1, 1))
                and (len(cast(GroupGrammar, subexpr).subexprs) == 1)
            ):
                subexpr = cast(GroupGrammar, subexpr).subexprs[0]
                wrap = isinstance(subexpr, GroupGrammar)
            return wrap
        return self.quantifier != (1, 1)

    @staticmethod
    def simplify_subexprs(
        original_subexprs: list[Grammar],
        quantifier: tuple[int, int | None],
    ) -> Grammar | None:
        subexprs: list[Grammar] = []
        n = 0
        for subexpr in original_subexprs:
            simplified = subexpr.simplify()
            if simplified is None:
                continue
            subexprs.append(simplified)
            n += 1

        # Empty expression is no-op
        if n < 1:
            return None

        # And that is (1, n) or optional (0, n) containing subexpressions that are all optional (0, n) and are equivalent to each other (excluding quantifier) can be simplified to a single subexpression with a quantifier
        # The upper bound of the quantifier is calculated by taking the sum of the upper bounds of the subexpressions and multiplying it by the upper bound of the outer expression
        if (
            quantifier[0] in (0, 1)
            and all(
                map(
                    lambda x: isinstance(x, GroupGrammar) and x.quantifier[0] == 0,
                    subexprs,
                )
            )
            and all(
                subexprs[i].equals(subexprs[i + 1], check_quantifier=False)
                for i in range(n - 1)
            )
        ):
            upper_bound = sum(
                map(
                    lambda x: cast(int, cast(GroupGrammar, x).quantifier[1]),
                    subexprs,
                )
            ) * cast(int, quantifier[1])
            return cast(GroupGrammar, subexprs[0]).simplify_subexprs(
                cast(GroupGrammar, subexprs[0]).subexprs,
                (0, upper_bound),
            )

        if n == 1:
            # Unwrap a single default (1, 1) subexpression
            if quantifier == (1, 1):
                return subexprs[0]
            # And that is optional (0, 1) can recursively unwrap to the first simple or non-single, non-default (1, 1), and non-optional (0, 1) grouped grammar.
            if (
                (quantifier == (0, 1))
                and isinstance(subexprs[0], GroupGrammar)
                and (subexprs[0].quantifier in ((0, 1), (1, 1)))
            ):
                return And.simplify_subexprs(subexprs[0].subexprs, quantifier)

        merged_n = merge_adjacent_default_and_grammars(subexprs, n)
        if merged_n < n:
            return And.simplify_subexprs(subexprs, quantifier)
        merged_n = merge_adjacent_string_grammars(subexprs, n)
        if merged_n < n:
            return And.simplify_subexprs(subexprs, quantifier)
        grouped, grouped_n = group_repeating_subexprs(subexprs, n)
        if grouped_n < n:
            return And.simplify_subexprs(grouped, quantifier)

        # And that is optional (0, n) containing subexpressions that are all optional (0, n) does not need to then be optional itself
        if (quantifier[0] == 0) and all(
            map(
                lambda x: isinstance(x, GroupGrammar) and (x.quantifier[0] == 0),
                subexprs,
            )
        ):
            return And.simplify_subexprs(subexprs, (1, quantifier[1]))

        return And(subexprs, quantifier=quantifier)


def merge_adjacent_default_and_grammars(subexprs: list[Grammar], n: int) -> int:
    """Merge adjacent And expressions having default quantifier (1, 1) in-place.

    Args:
        subexprs (list[Grammar]): Subexpressions to merge.
        n (int): Number of subexpressions.

    Returns:
        int: Number of subexpressions after merging.
    """
    if n < 2:
        return n
    last_idx = -1
    for i in range(n - 1, -1, -1):
        if isinstance(subexprs[i], And) and (
            cast(And, subexprs[i]).quantifier == (1, 1)
        ):
            if last_idx < 1:
                last_idx = i
            else:
                n -= 1
        else:
            if last_idx > 0:
                subexprs[i + 1] = And(
                    (
                        s
                        for subexpr in subexprs[i + 1 : last_idx + 1]
                        for s in cast(And, subexpr).subexprs
                    )
                )
                del subexprs[i + 2 : last_idx + 1]
            last_idx = -1
    if last_idx > 0:
        subexprs[0] = And(
            (
                s
                for subexpr in subexprs[0 : last_idx + 1]
                for s in cast(And, subexpr).subexprs
            )
        )
        del subexprs[1 : last_idx + 1]
    return n


def _iter_group_spans(n: int, size: int, offset: int) -> Iterator[tuple[int, int]]:
    """Generate all possible index ranges that would produce slices of the provided size, starting at the provided offset.

    Args:
        n (int): Number of subexpressions.
        size (int): Size of the group.
        offset (int): Offset to start from.

    Yields:
        tuple[int, int]: Start and end indices of the group.
            The range is inclusive, meaning both start and end indices are included.
            For example, (0, 2) includes indices 0, 1, and 2.
    """
    for end in range(n - offset, size - 2, -size):
        start = end - size
        if start < 0:
            return
        yield start, end


def group_repeating_subexprs(
    subexprs: list[Grammar],
    n: int,
) -> tuple[list[Grammar], int]:
    """Group repeating sequences of subexpressions into a single subexpression.

    Todo:

    * Split up strings into smaller strings when performing grouping (`.explode()`).
    * Improve the features used to sort so that it takes the minimal number of grammars when counted recursively to the lowest non-group grammar types.
      This function would likely then call itself and/or simplify. Be sure to avoid grouping repeating characters in a string if lower than a given threshold.

    Args:
        subexprs (list[Grammar]): Subexpressions to group.
        n (int): Length of the expression.

    Returns:
        tuple[list[Grammar], int]: Grouped subexpressions and the number of subexpressions after grouping.
    """
    best_subexprs = []
    best_n = 0
    best_weight = GroupWeight(0, 0, 0)
    max_chunk_size = n // 2
    for chunk_size in range(max_chunk_size, 0, -1):
        max_offset_a = n - (chunk_size * 2)
        max_offset_b = chunk_size - (1 + (n % chunk_size))
        max_offset = max(max_offset_a, max_offset_b)
        for offset in range(max_offset + 1):
            count = 0
            cmp_slice = []
            start = 0  # Prevent unbounded warning
            for start, end in _iter_group_spans(n, chunk_size, offset):
                assert end - start == chunk_size
                # Record initial slice to compare against
                if count == 0:
                    cmp_slice = subexprs[start:end]
                    count += 1
                    continue

                # Record additional matching slice
                if cmp_slice == subexprs[start:end]:
                    count += 1
                    continue

                # Yield variation of subexpressions where repeating slices are grouped inside a grammar
                if count > 1:
                    new_n = n - (count * chunk_size) + 1
                    weight = GroupWeight(n - new_n, chunk_size, count)
                    if weight > best_weight:
                        grouped_grammar = And(
                            subexprs[start : start + chunk_size],
                            quantifier=(count, count),
                        )
                        best_subexprs = (
                            subexprs[:start]
                            + [grouped_grammar]
                            + subexprs[start + (count * chunk_size) :]
                        )
                        best_n = new_n
                        best_weight = weight

                # Reset comparison slice
                cmp_slice = subexprs[start:end]
                count = 1

            if count > 1:
                new_n = n - (count * chunk_size) + 1
                weight = GroupWeight(n - new_n, chunk_size, count)
                if weight > best_weight:
                    grouped_grammar = And(
                        subexprs[start : start + chunk_size],
                        quantifier=(count, count),
                    )
                    best_subexprs = (
                        subexprs[:start]
                        + [grouped_grammar]
                        + subexprs[start + (count * chunk_size) :]
                    )
                    best_n = new_n
                    best_weight = weight

    if best_weight > GroupWeight(0, 0, 0):
        assert len(best_subexprs) > 0
        assert best_n > 0
        return best_subexprs, best_n
    assert len(best_subexprs) == 0
    assert best_n == 0
    return subexprs, n



"""Base grammar classes and utility functions for handling grouped grammar expressions.

Provides abstractions for building and rendering grouped grammar expressions, as well as
utility functions for grouping and merging subexpressions.
"""

from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from grammatica.grammar.base import Grammar

if sys.version_info >= (3, 12):  # pragma: no cover
    from typing import override
else:  # pragma: no cover
    from typing_extensions import override


if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any


class GroupGrammar(Grammar, ABC):
    """Base class for grouped grammar expressions.

    Args:
        subexprs (Iterable[Grammar]): Group of grammars.
        quantifier (int | tuple[int, int | None]): Minimum and maximum repetitions the expression must match.

    Raises:
        ValueError: Range lower bound is negative.
        ValueError: Range upper bound is not positive or None (infinity).
        ValueError: Range lower bound is greater than range upper bound.
    """

    __slots__: tuple[str, ...] = ("separator", "subexprs", "quantifier")

    separator: str
    """Separator to use for the grammar."""

    def __init__(
        self,
        *,
        subexprs: Iterable[Grammar],
        quantifier: int | tuple[int, int | None],
    ) -> None:
        super().__init__()

        self.subexprs: list[Grammar] = list(subexprs)
        """Group of grammars."""

        lower: int
        upper: int | None
        if isinstance(quantifier, int):
            lower = upper = quantifier
        else:  # tuple
            lower, upper = quantifier
        if lower < 0:
            raise ValueError(f"Range lower bound must be non-negative: {quantifier}")
        if upper is not None:
            if upper < 1:
                raise ValueError(
                    f"Range upper bound must be positive or None (infinity): {quantifier}"
                )
            if lower > upper:
                raise ValueError(
                    f"Range lower bound must be <= range upper bound: {quantifier}"
                )
        self.quantifier: tuple[int, int | None] = (lower, upper)
        """Minimum and maximum repetitions the expression must match."""

    def render(self, full: bool = True, wrap: bool = True) -> str | None:
        if len(self.subexprs) < 1:
            return None
        rendered_quantifier = self.render_quantifier()
        expr = ""
        found = False
        for subexpr in self.subexprs:
            rendered = subexpr.render(full=False)
            if rendered is not None:
                if found:
                    expr += self.separator
                expr += rendered
                found = True
        if not found:
            return None
        if self.needs_wrapped() and (wrap or (rendered_quantifier is not None)):
            expr = "(" + expr + ")"
        if rendered_quantifier is not None:
            expr += rendered_quantifier
        return expr

    def render_quantifier(self) -> str | None:
        """Render the quantifier.

        Returns:
            str | None: A quantifier string or None if not applicable.
        """
        lower, upper = self.quantifier
        if lower == upper == 1:
            return None
        if lower == 0:
            if upper is None:
                return "*"
            if upper == 1:
                return "?"
            # NOTE: Not supported by llama.cpp
            # return "{," + str(upper) + "}"
        if upper is None:
            if lower == 1:
                return "+"
            return "{" + str(lower) + ",}"
        if lower == upper:
            return "{" + str(lower) + "}"
        return "{" + str(lower) + "," + str(upper) + "}"

    def simplify(self) -> Grammar | None:
        return self.simplify_subexprs(self.subexprs, self.quantifier)

    @staticmethod
    @abstractmethod
    def simplify_subexprs(
        original_subexprs: list[Grammar],
        quantifier: tuple[int, int | None],
    ) -> Grammar | None:
        """Simplify the provided subexpressions for the grouped grammar.

        Args:
            original_subexprs (list[Grammar]): Subexpressions to simplify.
            quantifier (tuple[int, int | None]): Quantifier for the expression.

        Returns:
            Grammar | None: Simplified expression. If the expression is empty, return None.
        """
        return None

    @abstractmethod
    def needs_wrapped(self) -> bool:
        """Check if the expression needs to be wrapped in parentheses.

        Returns:
            bool: True if the expression needs to be wrapped, False otherwise.
        """
        return False

    @override
    def attrs_dict(self) -> dict[str, Any]:
        return {
            "subexprs": self.subexprs,
            "quantifier": self.quantifier,
        } | super().attrs_dict()
